###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY, PROP_INCONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.min_eq_propagator import compute_domains_min_eq


class TestMinEQ:
    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), (2, 5), (2, 6)])
        data = new_parameters_by_values([])
        assert compute_domains_min_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[2, 4], [2, 5], [2, 4]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(1, 3), (3, 3), (4, 5)])
        data = new_parameters_by_values([])
        assert compute_domains_min_eq(domains, data) == PROP_INCONSISTENCY

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(2, 4), (2, 5), (6, 8)])
        data = new_parameters_by_values([])
        assert compute_domains_min_eq(domains, data) == PROP_INCONSISTENCY

    def test_compute_domains_4(self) -> None:
        domains = new_shr_domains_by_values([(0, 1), (0, 1), (1, 1)])
        data = new_parameters_by_values([])
        assert compute_domains_min_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 1], [1, 1], [1, 1]]))

    def test_compute_domains_5(self) -> None:
        domains = new_shr_domains_by_values([(2, 3), (2, 3), (0, 1)])
        data = new_parameters_by_values([])
        assert compute_domains_min_eq(domains, data) == PROP_INCONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    DOM_UPDATE_EVENTS,
    DOM_UPDATE_IDX,
    EVENT_MASK_MAX,
    EVENT_MASK_MIN,
    EVENT_MASK_MIN_GROUND,
    MAX,
    MIN,
)
from nucs.solvers.choice_points import cp_put


@njit(cache=True)
def split_low_dom_heuristic(
    params: NDArray,
    shr_domains_stack: NDArray,
    not_entailed_propagators_stack: NDArray,
    dom_update_stack: NDArray,
    stacks_top: NDArray,
    dom_idx: int,
) -> int:
    """
    Chooses the first half of the domain.
    :param params: a two-dimensional parameters array, unused here
    :param shr_domains_stack: the stack of shared domains
    :param not_entailed_propagators_stack: the stack of not entailed propagators
    :param dom_update_stack: the stack of domain updates
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param dom_idx: the index of the shared domain
    :return: the events
    """
    cp_cur_idx = stacks_top[0]
    value = (shr_domains_stack[cp_cur_idx, dom_idx, MIN] + shr_domains_stack[cp_cur_idx, dom_idx, MAX]) // 2
    cp_put(shr_domains_stack, not_entailed_propagators_stack, stacks_top)
    shr_domains_stack[cp_cur_idx + 1, dom_idx, MAX] = value
    shr_domains_stack[cp_cur_idx, dom_idx, MIN] = value + 1
    dom_update_stack[cp_cur_idx, DOM_UPDATE_IDX] = dom_idx
    dom_update_stack[cp_cur_idx, DOM_UPDATE_EVENTS] = (
        EVENT_MASK_MIN_GROUND
        if shr_domains_stack[cp_cur_idx, dom_idx, MIN] == shr_domains_stack[cp_cur_idx, dom_idx, MAX]
        else EVENT_MASK_MIN
    )
    return EVENT_MASK_MAX



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from nucs.constants import STATS_IDX_SOLVER_SOLUTION_NB
from nucs.examples.tsp.tsp_problem import TSPProblem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestTSP:
    def test_tsp_1(self) -> None:
        problem = TSPProblem([[0, 2, 1, 2], [2, 0, 2, 1], [1, 2, 0, 2], [2, 1, 2, 0]])
        solver = BacktrackSolver(problem, decision_domains=[0, 1, 2, 3])
        solution = solver.minimize(problem.shr_domain_nb - 1)
        assert solution is not None
        assert solution[:4].tolist() == [1, 3, 0, 2]
        assert solution[problem.shr_domain_nb - 1] == 6
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 2

    # def test_tsp_gr17(self) -> None:
    #     problem = TSPProblem(GR17)
    #     solver = BacktrackSolver(
    #         problem, var_heuristic_idx=VAR_HEURISTIC_MAX_REGRET, dom_heuristic_idx=DOM_HEURISTIC_MIN_COST
    #     )
    #     solution = solver.minimize(problem.variable_nb - 1)
    #     assert solution is not None
    #     assert solution[problem.variable_nb - 1] == 2085

    # def test_tsp_gr21(self) -> None:
    #     problem = TSPProblem(GR21)
    #     solver = BacktrackSolver(problem)
    #     solution = solver.minimize(problem.variable_nb - 1)
    #     assert solution[problem.variable_nb - 1] == 2707
    #
    # def test_tsp_gr24(self) -> None:
    #     problem = TSPProblem(GR24)
    #     solver = BacktrackSolver(problem)
    #     solution = solver.minimize(problem.variable_nb - 1)
    #     assert solution[problem.variable_nb - 1] == 1272
